// ----------------------------------------------------------------------------------------------------------
// NEOCDWND.CPP
#include "burner.h"
#include "png.h"
#include "neocdlist.h"

int				NeoCDList_Init();
static void		NeoCDList_InitListView();
static int		NeoCDList_CheckISO(HWND hList, TCHAR* pszFile);
static void		NeoCDList_iso9660_CheckDirRecord(HWND hList, TCHAR* pszFile,  FILE* fp, int nSector);
static int		NeoCDList_AddGame(TCHAR* pszFile, unsigned int nGameID);
static void		NeoCDList_ScanDir(HWND hList, TCHAR* pszDirectory);
static TCHAR*	NeoCDList_ParseCUE(TCHAR* pszFile);
static int		NeoCDList_SelectDirectory(int nType);
static void		NeoCDList_ShowPreview(HWND hDlg, TCHAR* szFile, int nCtrID, int nFrameCtrID, float maxw, float maxh);
struct PNGRESOLUTION { int nWidth; int nHeight; };
static PNGRESOLUTION GetPNGResolution(TCHAR* szFile);

static BOOL	CALLBACK	NeoCDList_WndProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam);
static DWORD WINAPI		NeoCDList_DoProc(LPVOID);

static HWND		hNeoCDWnd			= NULL;
static HWND		hListView			= NULL;
static HWND		hProcParent			= NULL;
static int		nProcess			= 0;
static bool		bProcessingList		= false;
static HANDLE	hProcessThread[30];
static DWORD	dwProcessThreadID[30];

bool bNeoCDListScanSub			= false;
bool bNeoCDListScanOnlyISO		= false;
TCHAR szNeoCDCoverDir[MAX_PATH] = _T("support\\neocdz\\");
TCHAR szNeoCDGamesDir[MAX_PATH] = _T("ROMs\\neocdz\\");

static int nSelectedItem = -1;

struct GAMELIST {
	unsigned int nID;
	bool	bFoundCUE;
	TCHAR	szPathCUE[256];
	TCHAR	szPath[256];
	TCHAR	szISOFile[256];
	int		nAudioTracks;
	TCHAR	szTracks[99][256];
};

GAMELIST ngcd_list[100];
int nListItems = 0;

// Add game to List
static int NeoCDList_AddGame(TCHAR* pszFile, unsigned int nGameID) 
{
	NGCDGAME* game;
	game = (NGCDGAME*)malloc(sizeof(NGCDGAME));
	memset(game, 0, sizeof(NGCDGAME));
	
	if(GetNeoGeoCDInfo(nGameID))
	{		
		memcpy(game, GetNeoGeoCDInfo(nGameID), sizeof(NGCDGAME));
		
		TCHAR szNGCDID[12];
		_stprintf(szNGCDID, _T("%04X"), nGameID);
		
		LVITEM lvi;
		ZeroMemory(&lvi, sizeof(lvi));

		// NGCD-ID [Insert]
		lvi.iImage			= 0;
		lvi.iItem			= ListView_GetItemCount(hListView) + 1;
		lvi.mask			= LVIF_TEXT | LVIF_PARAM | LVIF_IMAGE;
		lvi.cchTextMax		= 256;
		TCHAR szTitle[256];
		_stprintf(szTitle, _T(" %s"), game->pszTitle);
		lvi.pszText			= szTitle;

		int nItem = ListView_InsertItem(hListView, &lvi);

		// TITLE
		ZeroMemory(&lvi, sizeof(lvi));

		lvi.iSubItem		= 1;
		lvi.iItem			= nItem;
		lvi.cchTextMax		= _tcslen(game->pszTitle);
		lvi.mask			= LVIF_TEXT | LVIF_IMAGE;
		lvi.pszText			= szNGCDID;

		ListView_SetItem(hListView, &lvi);

		// YEAR
		ZeroMemory(&lvi, sizeof(lvi));

		lvi.iSubItem		= 2;
		lvi.iItem			= nItem;
		lvi.cchTextMax		= _tcslen(game->pszYear);
		lvi.mask			= LVIF_TEXT;
		lvi.pszText			= game->pszYear;

		ListView_SetItem(hListView, &lvi);

		// COMPANY
		ZeroMemory(&lvi, sizeof(lvi));

		lvi.iSubItem		= 3;
		lvi.iItem			= nItem;
		lvi.cchTextMax		= _tcslen(game->pszCompany);
		lvi.mask			= LVIF_TEXT;
		lvi.pszText			= game->pszCompany;

		ListView_SetItem(hListView, &lvi);

		// SHORTNAME
		ZeroMemory(&lvi, sizeof(lvi));

		lvi.iSubItem		= 4;
		lvi.iItem			= nItem;
		lvi.cchTextMax		= _tcslen(game->pszCompany);
		lvi.mask			= LVIF_TEXT;
		lvi.pszText			= game->pszName;

		ListView_SetItem(hListView, &lvi);

		ngcd_list[nListItems].nID = nGameID;
		
		_tcscpy(ngcd_list[nListItems].szPath, pszFile);
		
		nListItems++;

	} else {
		// error
		return 0;
	}

	if(game) {
		free(game);
		game = NULL;
	}
	return 1;
}

static void NeoCDList_InitListView()
{
	LVCOLUMN LvCol;
	ListView_SetExtendedListViewStyle(hListView, LVS_EX_FULLROWSELECT);

	memset(&LvCol, 0, sizeof(LvCol));
	LvCol.mask		= LVCF_TEXT | LVCF_WIDTH | LVCF_SUBITEM;

	LvCol.cx		= 350;
	LvCol.pszText	= _T("Title");
	SendMessage(hListView, LVM_INSERTCOLUMN , 0, (LPARAM)&LvCol);

	LvCol.cx		= 70;
	LvCol.pszText	= _T("NGCD-ID");
	SendMessage(hListView, LVM_INSERTCOLUMN , 1, (LPARAM)&LvCol);

	LvCol.cx		= 50;
	LvCol.pszText	= _T("Year");
	SendMessage(hListView, LVM_INSERTCOLUMN , 2, (LPARAM)&LvCol);

	LvCol.cx		= 85;
	LvCol.pszText	= _T("Company");
	SendMessage(hListView, LVM_INSERTCOLUMN , 3, (LPARAM)&LvCol);

	LvCol.cx		= 85;
	LvCol.pszText	= _T("Shortname");
	SendMessage(hListView, LVM_INSERTCOLUMN , 4, (LPARAM)&LvCol);

	// Setup ListView Icons
	HIMAGELIST hImageList = ImageList_Create(16, 16, ILC_MASK | ILC_COLOR16, 0, 1);		
	ListView_SetImageList(hListView, hImageList, LVSIL_SMALL);
	ImageList_AddMasked(hImageList, LoadBitmap(hAppInst, MAKEINTRESOURCE(BMP_CD)), RGB(255, 0, 255));
}

static int NeoCDList_CheckDuplicates(HWND hList, unsigned int nID) 
{
	int nItemCount = ListView_GetItemCount(hList);

	for(int nItem = 0; nItem < nItemCount; nItem++) 
	{
		unsigned int nItemVal = 0;
		TCHAR szText[] = _T("0000");

		ListView_GetItemText(hList, nItem, 1, szText, 5);

		_stscanf(szText, _T("%x"), &nItemVal);

		if(nItemVal == nID) {
			ngcd_list[nItemCount].nAudioTracks = 0;
			return 1; // let's get out of here...
		}
	}

	return 0;
}

static void NeoCDList_iso9660_CheckDirRecord(HWND hList, TCHAR* pszFile,  FILE* fp, int nSector) 
{
	int		nSectorLength		= 2048;	
	//int		nFile				= 0;
	long	lBytesRead			= 0;
	//int		nLen				= 0;
	long	lOffset				= 0;
	bool	bNewSector			= false;
	bool	bRevisionQueve		= false;
	int		nRevisionQueveID	= 0;	

	lOffset = (nSector * nSectorLength);	

	while(1) 
	{
		unsigned char nLenDR;
		memcpy(&nLenDR, iso9660_ReadOffset(fp, lOffset, 1), 1);

		if(nLenDR == 0x22) {
			lOffset		+= nLenDR;
			lBytesRead	+= nLenDR;
			continue;
		}

		if(nLenDR < 0x22) 
		{
			if(bNewSector) 
			{
				if(bRevisionQueve) {
					bRevisionQueve		= false;

					// make sure we don't add duplicates to the list
					if(NeoCDList_CheckDuplicates(hList, nRevisionQueveID)) {
						return;
					}

					NeoCDList_AddGame(pszFile, nRevisionQueveID);
				}
				return;
			}

			nLenDR = 0;
			memcpy(&nLenDR, iso9660_ReadOffset(fp, lOffset + 1, 1), 1);

			if(nLenDR < 0x22) {
				lOffset += (2048 - lBytesRead);
				lBytesRead = 0;
				bNewSector = true;
				continue;
			}
		}

		bNewSector = false;

		unsigned char flags;
		memcpy(&flags, iso9660_ReadOffset(fp, lOffset + 25, 1), 1);

		if(!(flags & (1 << 1))) 
		{
			//MessageBoxA(NULL, "file", "", MB_OK);	

			unsigned char extentloc[8];
			memcpy(&extentloc, iso9660_ReadOffset(fp, lOffset + 2, 8), 8);

			char szValue[9];
			sprintf(szValue, "%02x%02x%02x%02x", extentloc[4], extentloc[5], extentloc[6], extentloc[7]);

			//MessageBoxA(NULL, szValue, "", MB_OK);

			unsigned int nValue = 0;
			sscanf(szValue, "%x", &nValue); 

			unsigned char data[0x10B];
			memcpy(&data, iso9660_ReadOffset(fp, (nValue * 2048), 0x10A), 0x10A);

			char szData[8];
			sprintf(szData, "%c%c%c%c%c%c%c", data[0x100], data[0x101], data[0x102], data[0x103], data[0x104], data[0x105], data[0x106]);

			if(!strncmp(szData, "NEO-GEO", 7)) 
			{
				_tcscpy(ngcd_list[nListItems].szISOFile, pszFile);

				//MessageBoxA(NULL, szData, "", MB_OK);

				char id[] = "0000";
				sprintf(id, "%02X%02X",  data[0x108], data[0x109]);

				unsigned int nID = 0;
				sscanf(id, "%x", &nID);

				unsigned char LEN_FI;
				memcpy(&LEN_FI, iso9660_ReadOffset(fp, lOffset + 32, 1), 1);

				char file[32];
				strncpy(file, (char*)iso9660_ReadOffset(fp, lOffset + 33, LEN_FI), LEN_FI);				
				file[LEN_FI] = 0;

				// King of Fighters '94, The (1994)(SNK)(JP)
				// 10-6-1994 (P1.PRG)
				if(nID == 0x0055 && data[0x67] == 0xDE) {
					// ...continue
				}

				
				// King of Fighters '94, The (1994)(SNK)(JP-US)
				// 11-21-1994 (P1.PRG)
				if(nID == 0x0055 && data[0x67] == 0xE6) {
					// Change to custom revision id
					nID = 0x1055;
				}
				
				
				// King of Fighters '95, The (1995)(SNK)(JP-US)[!][NGCD-084 MT B01, B03-B06, NGCD-084E MT B01]
				// 9-11-1995 (P1.PRG)
				if(nID == 0x0084 && data[0x6C] == 0xC0) {
					// ... continue
				}

				
				// King of Fighters '95, The (1995)(SNK)(JP-US)[!][NGCD-084 MT B10, NGCD-084E MT B03]
				// 10-5-1995 (P1.PRG)
				if(nID == 0x0084 && data[0x6C] == 0xFF) {
					// Change to custom revision id
					nID = 0x1084;
				}

				// King of Fighters '96 NEOGEO Collection, The
				if(nID == 0x0229) {
					bRevisionQueve = false;

					// make sure we don't add duplicates to the list
					if(NeoCDList_CheckDuplicates(hList, nID)) {
						return;
					}

					NeoCDList_AddGame(pszFile, nID);
					break;
				}

				// King of Fighters '96, The
				if(nID == 0x0214) {
					bRevisionQueve		= true;
					nRevisionQueveID	= nID;
					lOffset		+= nLenDR;
					lBytesRead	+= nLenDR;
					continue;
				}

				// make sure we don't add duplicates to the list
				if(NeoCDList_CheckDuplicates(hList, nID)) {
					return;
				}

				NeoCDList_AddGame(pszFile, nID);

				//MessageBoxA(NULL, id, "", MB_OK);
				break;
			}
		}	
		
		lOffset		+= nLenDR;
		lBytesRead	+= nLenDR;
	}
	return;
}

// Check the specified ISO, and proceed accordingly
static int NeoCDList_CheckISO(HWND hList, TCHAR* pszFile)
{
	if(!pszFile) {
		// error
		return 0;
	}

	// Make sure we have a valid ISO file extension...
	if(_tcsstr(pszFile, _T(".iso")) || _tcsstr(pszFile, _T(".ISO")) ) 
	{
		FILE* fp = _tfopen(pszFile, _T("rb"));
		if(fp) 
		{
			// Read ISO and look for 68K ROM standard program header, ID is always there
			// Note: This function works very quick, doesn't compromise performance :)
			// it just read each sector first 264 bytes aproximately only.

			// Get ISO Size (bytes)
			fseek(fp, 0, SEEK_END);
			long lSize = 0;
			lSize = ftell(fp);
			rewind(fp);

			iso9660_PVD pvd;

			// If it has at least 16 sectors proceed
			if(lSize > (2048 * 16)) 
			{	
				// Check for Standard ISO9660 Identifier
				unsigned char IsoHeader[2048 * 16 + 1];
				unsigned char IsoCheck[6];
		
				fread(IsoHeader, 1, 2048 * 16 + 1, fp); // advance to sector 16 and PVD Field 2
				fread(IsoCheck, 1, 5, fp);	// get Standard Identifier Field from PVD
		
				// Verify that we have indeed a valid ISO9660 MODE1/2048
				if(!memcmp(IsoCheck, "CD001", 5))
				{
					//bprintf(PRINT_NORMAL, _T("    Standard ISO9660 Identifier Found. \n"));
					iso9660_VDH vdh;

					// Get Volume Descriptor Header			
					memset(&vdh, 0, sizeof(vdh));
					memcpy(&vdh, iso9660_ReadOffset(fp, (2048 * 16), sizeof(vdh)), sizeof(vdh));

					// Check for a valid Volume Descriptor Type
					if(vdh.vdtype == 0x01) 
					{
						// Get Primary Volume Descriptor
						memset(&pvd, 0, sizeof(pvd));
						memcpy(&pvd, iso9660_ReadOffset(fp, (2048 * 16), sizeof(pvd)), sizeof(pvd));

						// ROOT DIRECTORY RECORD

						// Handle Path Table Location
						char szRootSector[32];
						unsigned int nRootSector = 0;

						sprintf(szRootSector, "%02X%02X%02X%02X", 
							pvd.root_directory_record.location_of_extent[4], 
							pvd.root_directory_record.location_of_extent[5], 
							pvd.root_directory_record.location_of_extent[6], 
							pvd.root_directory_record.location_of_extent[7]);

						// Convert HEX string to Decimal
						sscanf(szRootSector, "%X", &nRootSector);

						// Process the Root Directory Records
						NeoCDList_iso9660_CheckDirRecord(hList, pszFile, fp, nRootSector);

						// Path Table Records are not processed, since NeoGeo CD should not have subdirectories
						// ...
					}
				} else {

					//bprintf(PRINT_NORMAL, _T("    Standard ISO9660 Identifier Not Found, cannot continue. \n"));
					return 0;
				}
			}
		} else {

			//bprintf(PRINT_NORMAL, _T("    Couldn't open %s \n"), GetIsoPath());
			return 0;
		}

		if(fp) fclose(fp);

	} else {

		//bprintf(PRINT_NORMAL, _T("    File doesn't have a valid ISO extension [ .iso / .ISO ] \n"));
		return 0;
	}

	return 1;
}

// Scan the specified directory for sub-directories that contain ISO / CUE files
static void NeoCDList_ScanDir(HWND hList, TCHAR* pszDirectory)
{
	bProcessingList = true;

	ListView_DeleteAllItems(hList);

	WIN32_FIND_DATA ffdDirectory;

	HANDLE hDirectory = NULL;
	memset(&ffdDirectory, 0, sizeof(WIN32_FIND_DATA));
		
	// Scan main dir for sub-directories
	TCHAR szSearch[2048] = _T("\0");

	_stprintf(szSearch, _T("%s*"), pszDirectory);
	
	hDirectory = FindFirstFile(szSearch, &ffdDirectory);
	
	if (hDirectory == INVALID_HANDLE_VALUE) {
		// error
	} else {
		
		do 
		{
			// DIRECTORY
			if((ffdDirectory.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
			{
				if(!_tcscmp(ffdDirectory.cFileName, _T(".")) || !_tcscmp(ffdDirectory.cFileName, _T("..")))
				{
					// lets ignore " . " and " .. "
					continue;
				}

				TCHAR* pszISO = NULL;
				pszISO = (TCHAR*)malloc(sizeof(TCHAR) * 512);
				
				bool bDone = false;

				WIN32_FIND_DATA ffdSubDirectory;

				HANDLE hSubDirectory = NULL;
				memset(&ffdSubDirectory, 0, sizeof(WIN32_FIND_DATA));

				TCHAR szSubSearch[512] = _T("\0");

				if(!bNeoCDListScanOnlyISO) 
				{

					// Scan sub-directory for CUE									
					_stprintf(szSubSearch, _T("%s%s\\*.cue"), pszDirectory, ffdDirectory.cFileName);

					hSubDirectory = FindFirstFile(szSubSearch, &ffdSubDirectory);
				
					if (hSubDirectory == INVALID_HANDLE_VALUE) {
						// error
					} else {
		
						do
						{
							// Sub-directories only
							if(!(ffdSubDirectory.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
							{
								// File is CUE
								if(_tcsstr(ffdSubDirectory.cFileName, _T(".cue")) || _tcsstr(ffdSubDirectory.cFileName, _T(".CUE")))
								{
									// Parse CUE
									TCHAR szParse[512] = _T("\0");				
									_stprintf(szParse, _T("%s%s\\%s"), pszDirectory, ffdDirectory.cFileName, ffdSubDirectory.cFileName);

									//MessageBox(NULL, szParse, _T(""), MB_OK);

									pszISO = NeoCDList_ParseCUE( szParse );

									TCHAR szISO[512] =_T("\0");
									_stprintf(szISO, _T("%s%s\\%s"), pszDirectory, ffdDirectory.cFileName,  pszISO);

									NeoCDList_CheckISO(hList, szISO);
									bDone = true;

									break; // no need to continue				
								}
							}
						} while(FindNextFile(hSubDirectory, &ffdSubDirectory));
					}

					if(pszISO) {
						pszISO[0] = 0;
					}

					if(bDone) {
						FindClose(hSubDirectory);
						continue;
					}
				} else {
					if(!ngcd_list[nListItems].nAudioTracks) {
						ngcd_list[nListItems].nAudioTracks = 0;
					}
				}

				hSubDirectory = NULL;
				memset(&ffdSubDirectory, 0, sizeof(WIN32_FIND_DATA));

				// Scan sub-directory for ISO
				_stprintf(szSubSearch, _T("%s%s\\*.iso"), pszDirectory, ffdDirectory.cFileName);

				hSubDirectory = FindFirstFile(szSubSearch, &ffdSubDirectory);

				if (hSubDirectory == INVALID_HANDLE_VALUE) {
					// error
				} else {
		
					do
					{
						// Sub-directories only
						if(!(ffdSubDirectory.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
						{
							// File is ISO
							if(_tcsstr(ffdSubDirectory.cFileName, _T(".iso")) || _tcsstr(ffdSubDirectory.cFileName, _T(".ISO")))
							{
								TCHAR szISO[512] = _T("\0");				
								_stprintf(szISO, _T("%s%s\\%s"), pszDirectory, ffdDirectory.cFileName, ffdSubDirectory.cFileName);

								NeoCDList_CheckISO(hList, szISO);

								break; // no need to continue
							}
						}
					} while(FindNextFile(hSubDirectory, &ffdSubDirectory));
				}
			} else {
				// FILE
			}
		} while(FindNextFile(hDirectory, &ffdDirectory));			
	}
	FindClose(hDirectory);
	
	bProcessingList = false;
}

// Scan the specified directory for ISO / CUE files
static void NeoCDList_ScanSingleDir(HWND hList, TCHAR* pszDirectory)
{
	bProcessingList = true;
	ListView_DeleteAllItems(hList);

	//	
	TCHAR* pszISO = NULL;
	pszISO = (TCHAR*)malloc(sizeof(TCHAR) * 512);
				
	WIN32_FIND_DATA ffdDirectory;

	HANDLE hDirectory = NULL;
	memset(&ffdDirectory, 0, sizeof(WIN32_FIND_DATA));

	// Scan directory for CUE
	TCHAR szSearch[512] = _T("\0");

	if(!bNeoCDListScanOnlyISO) 
	{
		_stprintf(szSearch, _T("%s*.cue"), pszDirectory);

		hDirectory = FindFirstFile(szSearch, &ffdDirectory);

		if (hDirectory == INVALID_HANDLE_VALUE) {
			// error
		} else {

			do
			{
				// Files only
				if(!(ffdDirectory.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
				{
					// File is CUE
					if(_tcsstr(ffdDirectory.cFileName, _T(".cue")) || _tcsstr(ffdDirectory.cFileName, _T(".CUE")))
					{
						// Parse CUE
						TCHAR szParse[512] = _T("\0");				
						_stprintf(szParse, _T("%s%s"), pszDirectory, ffdDirectory.cFileName);

						//MessageBox(NULL, szParse, _T(""), MB_OK);

						pszISO = NeoCDList_ParseCUE( szParse );

						TCHAR szISO[512] =_T("\0");
						_stprintf(szISO, _T("%s%s"), pszDirectory, pszISO);

						NeoCDList_CheckISO(hList, szISO);
					}
				}
			} while(FindNextFile(hDirectory, &ffdDirectory));

			FindClose(hDirectory);
		}
	} else {
		if(!ngcd_list[nListItems].nAudioTracks) {
			ngcd_list[nListItems].nAudioTracks = 0;
		}
	}

	hDirectory = NULL;
	memset(&ffdDirectory, 0, sizeof(WIN32_FIND_DATA));

	// Scan directory for ISO
	_stprintf(szSearch, _T("%s*.iso"), pszDirectory);

	hDirectory = FindFirstFile(szSearch, &ffdDirectory);

	if (hDirectory == INVALID_HANDLE_VALUE) {
		// error
	} else {
		
		do
		{
			// Files only
			if(!(ffdDirectory.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
			{
				// File is ISO
				if(_tcsstr(ffdDirectory.cFileName, _T(".iso")) || _tcsstr(ffdDirectory.cFileName, _T(".ISO")))
				{
					TCHAR szISO[512] = _T("\0");				
					_stprintf(szISO, _T("%s%s"), pszDirectory, ffdDirectory.cFileName);

					NeoCDList_CheckISO(hList, szISO);

				}
			}
		} while(FindNextFile(hDirectory, &ffdDirectory));

		FindClose(hDirectory);
	}

	bProcessingList = false;
}


// This will parse the specified CUE file and return the ISO path, if found
static TCHAR* NeoCDList_ParseCUE(TCHAR* pszFile) 
{
	//if(!pszFile) return NULL;

	TCHAR* szISO = NULL;
	szISO = (TCHAR*)malloc(sizeof(TCHAR) * 2048);
	if(!szISO) return NULL;

	// open file
	FILE* fp = NULL;
	fp = _tfopen(pszFile, _T("r"));

	if(!fp) {
		return NULL;
	}

	while(!feof(fp))
	{
		TCHAR szBuffer[2048];
		TCHAR szOriginal[2048];

		_fgetts(szBuffer, sizeof(szBuffer), fp);

		// terminate string
		szBuffer[260] = 0;

		if(!*szBuffer) {
			return NULL;
		}

		int nLength = 0;
		nLength = _tcslen(szBuffer);

		// Remove ASCII control characters from the string (including the 'space' character)
		while (szBuffer[nLength-1] < 32 && nLength > 0)
		{
			szBuffer[nLength-1] = 0;
			nLength--;
		}

		_tcscpy(szOriginal, szBuffer);

		if(!_tcsncmp(szBuffer, _T("FILE"), 4))
		{
			TCHAR* pEnd = _tcsrchr(szBuffer, '"');
			if (!pEnd)	{
				break;	// Invalid CUE format
			}

			*pEnd = 0;

			TCHAR* pStart = _tcschr(szBuffer, '"');

			if(!pStart)	{
				break;	// Invalid CUE format
			}

			if(!_tcsncmp(pEnd + 2, _T("BINARY"), 6))
			{
				_tcscpy(szISO,  pStart + 1);
				ngcd_list[nListItems].bFoundCUE = true;
				_tcscpy(ngcd_list[nListItems].szPathCUE,  pszFile);
				_tcscpy(ngcd_list[nListItems].szISOFile,  pStart + 1);
			} 
			
			if (!_tcsncmp(pEnd + 2, _T("WAVE"), 5)) {
				if(!ngcd_list[nListItems].nAudioTracks) {
					ngcd_list[nListItems].nAudioTracks = 0;
				}

				_tcscpy(ngcd_list[nListItems].szTracks[ngcd_list[nListItems].nAudioTracks], pStart + 1);

				ngcd_list[nListItems].nAudioTracks++;
			} 
		}
	}
	if(fp) fclose(fp);

	return szISO;
}

static PNGRESOLUTION GetPNGResolution(TCHAR* szFile)
{
	int width = 0;
	int height = 0;
	PNGRESOLUTION nResolution = { 0, 0 };
	png_structp png_ptr;
	png_infop info_ptr;
	char header[8];

	FILE *fp = _tfopen(szFile, _T("rb"));

	if (!fp) {
		return nResolution;
	}

	fread(header, 1, 8, fp);

	if (png_sig_cmp((png_const_bytep)header, 0, 8)) {
		return nResolution;
	}

	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

	if (!png_ptr) {
		return nResolution;
	}

	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr) {
		return nResolution;
	}

	if (setjmp(png_jmpbuf(png_ptr))) {
		return nResolution;
	}

	png_init_io(png_ptr, fp);
	png_set_sig_bytes(png_ptr, 8);
	png_read_info(png_ptr, info_ptr);

	width	= png_get_image_width(png_ptr, info_ptr);
	height	= png_get_image_height(png_ptr, info_ptr);

	nResolution.nWidth = width;
	nResolution.nHeight = height;

	fclose(fp);

	return nResolution;
}

static void NeoCDList_ShowPreview(HWND hDlg, TCHAR* szFile, int nCtrID, int nFrameCtrID, float maxw, float maxh) 
{
	PNGRESOLUTION PNGRes = { 0, 0 };
	if(!_tfopen(szFile, _T("rb"))) 
	{
		HRSRC hrsrc			= FindResource(NULL, MAKEINTRESOURCE(BMP_SPLASH), RT_BITMAP);
		HGLOBAL hglobal		= LoadResource(NULL, (HRSRC)hrsrc);

		BITMAPINFOHEADER* pbmih = (BITMAPINFOHEADER*)LockResource(hglobal);

		PNGRes.nWidth	= pbmih->biWidth;
		PNGRes.nHeight	= pbmih->biHeight;

		FreeResource(hglobal);

	} else {
		PNGRes = GetPNGResolution(szFile);
	}

	// ------------------------------------------------------
	// PROPER ASPECT RATIO CALCULATIONS 

	float w = (float)PNGRes.nWidth;
	float h = (float)PNGRes.nHeight;

	//float maxw = 216; // 
	//float maxh = 150; // 

	// max WIDTH
	if(w > maxw) {
		float nh = maxw * (float)(h / w);
		float nw = maxw;

		// max HEIGHT
		if(nh > maxh) {
			nw = maxh * (float)(nw / nh);
			nh = maxh;
		}

		w = nw;
		h = nh;
	}

	// max HEIGHT
	if(h > maxh) {
		float nw = maxh * (float)(w / h);
		float nh = maxh;

		// max WIDTH
		if(nw > maxw) {
			nh = maxw * (float)(nh / nw);
			nw = maxw;
		}

		w = nw;
		h = nh;
	}

	// Proper centering of preview
	float x = ((maxw - w) / 2);
	float y = ((maxh - h) / 2);

	RECT rc = {0, 0, 0, 0};
	GetWindowRect(GetDlgItem(hDlg, nFrameCtrID), &rc);

	POINT pt;
	pt.x = rc.left;
	pt.y = rc.top;

	ScreenToClient(hDlg, &pt);

	// ------------------------------------------------------

	FILE* fp = _tfopen(szFile, _T("rb"));
	
	HBITMAP hCoverBmp = PNGLoadBitmap(hDlg, fp, (int)w, (int)h, 0);

	SetWindowPos(GetDlgItem(hDlg, nCtrID), NULL, (int)(pt.x + x), (int)(pt.y + y), 0, 0, SWP_NOSIZE | SWP_SHOWWINDOW);
	SendDlgItemMessage(hDlg, nCtrID, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hCoverBmp);

	if(fp) fclose(fp);

}

static void NeoCDList_Clean() 
{
	NeoCDList_ShowPreview(hNeoCDWnd, _T(""), IDC_NCD_FRONT_PIC, IDC_NCD_FRONT_PIC_FRAME, 216, 150);
	NeoCDList_ShowPreview(hNeoCDWnd, _T(""), IDC_NCD_BACK_PIC, IDC_NCD_BACK_PIC_FRAME, 216, 150);

	SetWindowText(GetDlgItem(hNeoCDWnd, IDC_NCD_CDIMG_EDIT), _T(""));
	SetWindowText(GetDlgItem(hNeoCDWnd, IDC_NCD_TRACKS_EDIT), _T(""));

	memset(&hProcessThread, 0, sizeof(HANDLE) * 30);
	memset(&dwProcessThreadID, 0, sizeof(HANDLE) * 30);
	for(int x = 0; x < 100; x++) {
		for(int y = 0; y < 99; y++) {
			memset(&ngcd_list[x].szTracks[y], 0, sizeof(TCHAR) * 256);
		}
		ngcd_list[x].bFoundCUE = false;
	}
	memset(&ngcd_list, 0, (sizeof(GAMELIST) * 100));

	nProcess			= 0;
	hProcParent			= NULL;
	bProcessingList		= false;

	nListItems = 0;

	nSelectedItem = -1;
}

static int NeoCDList_SelectDirectory(int nType) {
	
	int nRet = 0;

	LPMALLOC pMalloc		= NULL;
	BROWSEINFO bInfo;
	ITEMIDLIST* pItemIDList = NULL;
	TCHAR buffer[1024];

	SHGetMalloc(&pMalloc);

	memset(&bInfo, 0, sizeof(bInfo));
	bInfo.hwndOwner			= hNeoCDWnd;
	bInfo.pszDisplayName	= buffer;
	bInfo.lpszTitle			= _T("Select Directory");
	bInfo.ulFlags			= BIF_EDITBOX | BIF_RETURNONLYFSDIRS | BIF_NEWDIALOGSTYLE;
	
	pItemIDList				= SHBrowseForFolder(&bInfo);

	if (pItemIDList) 
	{
		// Get the working directory [fix]
		TCHAR szWorkingDir[1024];
		GetCurrentDirectory(sizeof(szWorkingDir), szWorkingDir);	

		BOOL bDirSel = SHGetPathFromIDList(pItemIDList, buffer);

		if(bDirSel) 
		{
			// Set the working directory [fix]
			SetCurrentDirectory(szWorkingDir);
			
			int strLen	= _tcslen(buffer);

			if (strLen) 
			{
				if (buffer[strLen - 1] != '\\') {
					buffer[strLen]		= '\\';
					buffer[strLen + 1]	= '\0';
				}

				// GAMES
				if(nType == 0) {
					_stprintf(szNeoCDGamesDir, _T("%s"), buffer);
				}

				// COVERS
				if(nType == 1) {
					_stprintf(szNeoCDCoverDir, _T("%s"), buffer);
				}
			}

			nRet = 1;
		}
		pMalloc->Free(pItemIDList); 
	}
	pMalloc->Release();		

	return nRet;
}

static HWND hNeoCDList_CoverDlg = NULL;
TCHAR szBigCover[512] = _T("");

static BOOL CALLBACK NeoCDList_CoverWndProc(HWND hDlg, UINT Msg, WPARAM /*wParam*/, LPARAM /*lParam*/)
{
	if(Msg == WM_INITDIALOG)
	{
		hNeoCDList_CoverDlg = hDlg;

		ShowWindow(hDlg, SW_SHOW);

		NeoCDList_ShowPreview(hDlg, szBigCover, IDC_NCD_COVER_PREVIEW_PIC, IDC_NCD_COVER_PREVIEW_PIC, 580, 415);

		return TRUE;
	}

	if(Msg == WM_CLOSE)
	{		
		EndDialog(hDlg, 0);
		hNeoCDList_CoverDlg	= NULL;
	}

	return 0;
}

void NeoCDList_InitCoverDlg()
{
	DialogBox(hAppInst, MAKEINTRESOURCE(IDD_NCD_COVER_DLG), hNeoCDWnd, NeoCDList_CoverWndProc);
}

static BOOL CALLBACK NeoCDList_WndProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	if(Msg == WM_INITDIALOG)
	{
		hNeoCDWnd = hDlg;

		InitCommonControls();

		hListView = GetDlgItem(hDlg, IDC_NCD_LIST);

		NeoCDList_InitListView();

		HICON hIcon = LoadIcon(hAppInst, MAKEINTRESOURCE(IDI_APP));
		SendMessage(hDlg, WM_SETICON, ICON_SMALL, (LPARAM)hIcon);		// Set the Game Selection dialog icon.

		ShowWindow(hDlg, SW_SHOW);
		
		NeoCDList_ShowPreview(hNeoCDWnd, _T(""), IDC_NCD_FRONT_PIC, IDC_NCD_FRONT_PIC_FRAME, 216, 150);
		NeoCDList_ShowPreview(hNeoCDWnd, _T(""), IDC_NCD_BACK_PIC, IDC_NCD_BACK_PIC_FRAME, 216, 150);

		SetWindowText(GetDlgItem(hNeoCDWnd, IDC_NCD_CDIMG_EDIT), _T(""));
		SetWindowText(GetDlgItem(hNeoCDWnd, IDC_NCD_TRACKS_EDIT), _T(""));

		CheckDlgButton(hNeoCDWnd, IDC_NCD_SSUBDIR_CHECK, bNeoCDListScanSub ? BST_CHECKED : BST_UNCHECKED);
		CheckDlgButton(hNeoCDWnd, IDC_NCD_SISO_ONLY_CHECK, bNeoCDListScanOnlyISO ? BST_CHECKED : BST_UNCHECKED);

		hProcessThread[0]	= CreateThread(NULL, 0, NeoCDList_DoProc, NULL, 0, &dwProcessThreadID[nProcess]);		
	
		return TRUE;
	}

	if(Msg == WM_CLOSE)
	{
		NeoCDList_Clean();

		hNeoCDWnd	= NULL;
		hListView	= NULL;

		EndDialog(hDlg, 0);
	}
	
	if (Msg == WM_NOTIFY) 
	{
		NMLISTVIEW* pNMLV	= (NMLISTVIEW*)lParam;

		// Game Selected
		if (pNMLV->hdr.code == LVN_ITEMCHANGED && pNMLV->hdr.idFrom == IDC_NCD_LIST) 
		{
			int iCount		= SendMessage(hListView, LVM_GETITEMCOUNT, 0, 0);
			int iSelCount	= SendMessage(hListView, LVM_GETSELECTEDCOUNT, 0, 0);

			if(iCount == 0 || iSelCount == 0) return 1;

			TCHAR szID[] = _T("0000");

			int iItem = pNMLV->iItem;

			LVITEM LvItem;
			memset(&LvItem, 0, sizeof(LvItem));

			LvItem.iItem		= iItem;
			LvItem.mask			= LVIF_TEXT;
			LvItem.iSubItem		= 1;				// id column
			LvItem.pszText		= szID;
			LvItem.cchTextMax	= sizeof(szID) + 1;

			SendMessage(hListView, LVM_GETITEMTEXT, (WPARAM)iItem, (LPARAM)&LvItem);

			//MessageBox(NULL, LvItem.pszText, _T(""), MB_OK);

			TCHAR szFront[512];
			TCHAR szBack[512];

			_stprintf(szFront, _T("%s%s-front.png"), szNeoCDCoverDir, LvItem.pszText );
			_stprintf(szBack, _T("%s%s-back.png"), szNeoCDCoverDir, LvItem.pszText );

			// Front / Back Cover preview
			NeoCDList_ShowPreview(hNeoCDWnd, szFront, IDC_NCD_FRONT_PIC, IDC_NCD_FRONT_PIC_FRAME, 216, 150);
			NeoCDList_ShowPreview(hNeoCDWnd, szBack, IDC_NCD_BACK_PIC, IDC_NCD_BACK_PIC_FRAME, 216, 150);

			// Disc Image / Audio Tracks
			SetWindowText(GetDlgItem(hNeoCDWnd, IDC_NCD_CDIMG_EDIT), _T(""));
			SetWindowText(GetDlgItem(hNeoCDWnd, IDC_NCD_TRACKS_EDIT), _T(""));

			for(int nItem = 0; nItem < nListItems; nItem++) 
			{
				unsigned int nID = 0;
				_stscanf(szID, _T("%x"), &nID);

				TCHAR szAudioTracks[] =_T("0");				

				if(nID == ngcd_list[nItem].nID) {
					SetWindowText(GetDlgItem(hNeoCDWnd, IDC_NCD_CDIMG_EDIT), ngcd_list[nItem].szISOFile);
					_stprintf(szAudioTracks, _T("%d"), ngcd_list[nItem].nAudioTracks);
					SetWindowText(GetDlgItem(hNeoCDWnd, IDC_NCD_TRACKS_EDIT), szAudioTracks);

					nSelectedItem = nItem;
					break;
				}
			}

			
		}

		// Double Click
		if (pNMLV->hdr.code == NM_DBLCLK && pNMLV->hdr.idFrom == IDC_NCD_LIST) 
		{
			if(nSelectedItem >= 0) {
				nCDEmuSelect = 0;
				if(ngcd_list[nSelectedItem].bFoundCUE) {
					_tcscpy(CDEmuImage, ngcd_list[nSelectedItem].szPathCUE);
				} else {
					_tcscpy(CDEmuImage, ngcd_list[nSelectedItem].szPath);
				}
			} else {
				return 0;
			}

			NeoCDList_Clean();

			hNeoCDWnd	= NULL;
			hListView	= NULL;					

			EndDialog(hDlg, 0);

			int BurnerLoadDriver(TCHAR *szDriverName);

			BurnerLoadDriver(_T("neocdz"));
		}
	}

	if(Msg == WM_COMMAND)
	{
		if(HIWORD(wParam) == STN_CLICKED) 
		{
			int nCtrlID = LOWORD(wParam);

			if(nCtrlID == IDC_NCD_FRONT_PIC) 
			{
				if(nSelectedItem >= 0) {
					_stprintf(szBigCover, _T("%s%04x-front.png"), szNeoCDCoverDir, ngcd_list[nSelectedItem].nID );

					if(!_tfopen(szBigCover, _T("rb"))) {
						szBigCover[0] = 0;
						return 0;
					}
				} else {
					return 0;
				}
				NeoCDList_InitCoverDlg();
				return 0;
			}

			if(nCtrlID == IDC_NCD_BACK_PIC) 
			{
				if(nSelectedItem >= 0) {
					_stprintf(szBigCover, _T("%s%04x-back.png"), szNeoCDCoverDir, ngcd_list[nSelectedItem].nID );

					if(!_tfopen(szBigCover, _T("rb"))) {
						szBigCover[0] = 0;
						return 0;
					}
				} else {
					return 0;
				}
				NeoCDList_InitCoverDlg();
				return 0;
			}
		}

		if(HIWORD(wParam) == BN_CLICKED) 
		{
			int nCtrlID	= LOWORD(wParam);

			switch(nCtrlID) 
			{
				case IDC_NCD_PLAY_BUTTON:
				{
					if(nSelectedItem >= 0) {
						nCDEmuSelect = 0;
						if(ngcd_list[nSelectedItem].bFoundCUE) {
							_tcscpy(CDEmuImage, ngcd_list[nSelectedItem].szPathCUE);
						} else {
							_tcscpy(CDEmuImage, ngcd_list[nSelectedItem].szPath);
						}
					} else {
						break;
					}

					NeoCDList_Clean();

					hNeoCDWnd	= NULL;
					hListView	= NULL;					

					EndDialog(hDlg, 0);

					int BurnerLoadDriver(TCHAR *szDriverName);

					BurnerLoadDriver(_T("neocdz"));
					break;
				}

				case IDC_NCD_SCAN_BUTTON:
				{
					if(bProcessingList) break;

					NeoCDList_Clean();
					hProcessThread[0] = CreateThread(NULL, 0, NeoCDList_DoProc, NULL, 0, &dwProcessThreadID[nProcess]);	
					break;
				}

				case IDC_NCD_SEL_GAMES_DIR_BUTTON:
				{
					NeoCDList_SelectDirectory(0);
					break;
				}

				case IDC_NCD_SEL_COVERS_DIR_BUTTON:
				{
					NeoCDList_SelectDirectory(1);
					break;
				}

				case IDC_NCD_LOAD_CDIMG_BUTTON:
				{
					nCDEmuSelect = 0;
					
					// Get the working directory [fix]
					TCHAR szWorkingDir[1024] = _T("\0");
					GetCurrentDirectory(sizeof(szWorkingDir), szWorkingDir);

					memset(&ofn, 0, sizeof(ofn));
					ofn.lStructSize		= sizeof(ofn);
					ofn.hwndOwner		= hScrnWnd;
					ofn.lpstrFile		= CDEmuImage;
					ofn.nMaxFile		= sizeof(CDEmuImage);
					ofn.lpstrTitle		= _T("Select CD Image");
					ofn.lpstrFilter		= _T("CD images (*.iso,*.cue)\0*.iso;*.cue\0\0)");
					ofn.lpstrInitialDir = _T(".");
					ofn.Flags			= OFN_NOCHANGEDIR | OFN_HIDEREADONLY;
					ofn.lpstrDefExt		= _T("iso");

					if (GetOpenFileName(&ofn))
					{
						NeoCDList_Clean();

						hNeoCDWnd	= NULL;
						hListView	= NULL;					

						EndDialog(hDlg, 0);

						// Set the working directory [fix]
						SetCurrentDirectory(szWorkingDir);

						int BurnerLoadDriver(TCHAR *szDriverName);

						BurnerLoadDriver(_T("neocdz"));
					}
					break;
				}

				case IDC_NCD_SSUBDIR_CHECK:
				{
					if(BST_CHECKED == IsDlgButtonChecked(hDlg, IDC_NCD_SSUBDIR_CHECK)) {
						bNeoCDListScanSub = true;
					} else {
						bNeoCDListScanSub = false;
					}
					break;
				}

				case IDC_NCD_SISO_ONLY_CHECK:
				{
					if(BST_CHECKED == IsDlgButtonChecked(hDlg, IDC_NCD_SISO_ONLY_CHECK)) {
						bNeoCDListScanOnlyISO = true;
					} else {
						bNeoCDListScanOnlyISO = false;
					}
					break;
				}

				case IDC_NCD_CANCEL_BUTTON:
				{
					NeoCDList_Clean();

					hNeoCDWnd	= NULL;
					hListView	= NULL;				

					EndDialog(hDlg, 0);
					break;
				}
			}
		}
	}
	return 0;
}

static DWORD WINAPI NeoCDList_DoProc(LPVOID) 
{
	if(bProcessingList) return 0;

	SetThreadPriority(hProcessThread[0], THREAD_PRIORITY_TIME_CRITICAL);

	if(bNeoCDListScanSub) {
		NeoCDList_ScanDir(hListView, szNeoCDGamesDir);	
	} else {
		NeoCDList_ScanSingleDir(hListView, szNeoCDGamesDir);
	}

	CloseHandle(hProcessThread[0]);
	
	return 0;
}

int NeoCDList_Init() 
{
	DialogBox(hAppInst, MAKEINTRESOURCE(IDD_NCD_DLG), hScrnWnd, NeoCDList_WndProc);
	return 0;
}
